module sq4d.class_;

import sq4d.all;
import std.traits;
import std.string;
import std.stdio;
import std.exception;
import core.memory;

/// Dのオブジェクトのポインタが格納されるスロット名
enum DObjectSlotname = "_d_ptr";


/// クラスをget
SquirrelObject sqd_getclass(T)(HSQUIRRELVM sq, int index)
    if (is(T : Object))
{
    mixin(sqd_assert_stack_is_unchanged);
    return sqd_get!SquirrelObject(sq, index);
}

/// クラスの実体をget
T sqd_getclassentity(T)(HSQUIRRELVM sq, int index)
    if (is(T : Object))
{
    mixin(sqd_assert_stack_is_unchanged);
    auto o = sqd_get!SquirrelObject(sq, index);
    return cast(T)o.get!SQUserPointer(DObjectSlotname);
}

/// Dのクラスをラップしてpush。
/// この関数はスタックを変動させる。
void sqd_pushclass(T)(HSQUIRRELVM sq, T value)
    if (is(T : Object))
{
    debug {
        auto top = sq_gettop(sq);
        scope(success) sqd_enforce(top+1 == sq_gettop(sq), sq, "stack destroyed");
    }

    sqd_classensuredefine!T(sq);

    // push class
    sq_pushroottable(sq);
    sqd_push(sq, T.stringof);
    sq_get(sq, -2);
    sq_remove(sq, -2); // roottable

    sq_pushroottable(sq); // self

    auto retval = sqd_call!SquirrelObject(sq, cast(void*)value); // do "return T(value)"
    sqd_push(sq, retval);
}



/// メソッドをラップ
private extern(C)
SQInteger sqd_methodwrapper(T, string method)(HSQUIRRELVM sq) {
    try {
        /*
        static bool matchParam(alias F)(HSQUIRRELVM sq, int top) {
            alias ParameterTypeTuple!F PT;
            if (staticLength!PT != top)
                return false;
            foreach (i, T; PT) {
                auto type = sq_gettype(sq, i+1);
                if (is(T)) {
                }
            }
        }
        */
        
        //writeln("sqd_methodwrapper!("~T.stringof~", "~method~")");
        //scope(exit) writeln("return");
        debug auto top = sq_gettop(sq);
        //alias __traits(getOverloads, T, method) FS;
        mixin("alias T."~method~" F;");
        alias ParameterTypeTuple!F PT;
        alias ReturnType!F RT;
        //argument
        //writeln("load stack");
        auto self_ = sqd_get!SquirrelObject(sq, 1);
        auto self = self_.dobject!T;
        //PT args; // <- bugs: PT.~this() is not called
        Tuple!PT args;
        foreach (i, v; PT) {
            args[i] = sqd_get!(typeof(args[i]))(sq, -PT.length + i);
        }
        //writeln("call return");
        //call&return
        //mixin PrintFieldAttribute!(T, method);
        static if (is(RT == void)) {
            mixin("self."~method~"(args.tupleof);");
            debug sqd_enforce(top == sq_gettop(sq), sq, "stack destroyed");
            return 0;
        }
        else {
            auto retval = mixin("self."~method~"(args.tupleof)");
            sqd_push(sq, retval);
            debug sqd_enforce(top+1 == sq_gettop(sq), sq, "stack destroyed");
            return 1;
        }
    }
    catch (Throwable o) {
        return sq_throwerror(sq, ("sqd_methodwrapper!("~T.stringof~", "~method~") caught error: "~o.toString()).toStringz());
    }
}

/// Dメソッドをラップしてpushする。
/// この関数はスタックを変動させる。
void sqd_pushmethodwrapper(T, string method)(HSQUIRRELVM sq) {
    mixin("alias T."~method~" F;");
    sq_newclosure(sq, &sqd_methodwrapper!(T, method), 0);
    enum schemer = "x" ~ SquirrelSchemer!(ParameterTypeTuple!F) ~ "\0";
    sqd_checkresult(sq_setparamscheck(sq, SQ_MATCHTYPEMASKSTRING, schemer), sq, "sq_setparamscheck!("~T.stringof~", "~method~")("~schemer~")");
}

/// セッターの定義
private extern(C)
SQInteger sqd_classfieldsetter(T)(HSQUIRRELVM sq) {
    try {
        auto self_ = sqd_get!SquirrelObject(sq, 1);
        auto slotname = sqd_get!SQDString(sq, 2);
        auto self = self_.dobject!T;
        
        pragma(msg, "make setter");
        foreach (i, field_; __traits(derivedMembers, T)) {
            enum field = __traits(derivedMembers, T)[i];
            //static if (isMemberVariable!(T, field)) {
            static if (isProperty!(T, field)) {
                pragma(msg, T.stringof~"."~field);
                if (slotname == field) {
                    mixin("alias typeof(T."~field~") fieldtype;");
                    auto setval = sqd_get!fieldtype(sq, 3);
                    __traits(getMember, self, field) = setval;
                    sq_push(sq, 3); // is this must?
                    return 1;
                }
            }
        }
        return sq_throwerror(sq, ("sqd_classfieldsetter!"~T.stringof~"() set undefined slot: "~T.stringof~"."~slotname).toStringz());
    }
    catch (Throwable o) {
        return sq_throwerror(sq, ("sqd_classfieldsetter!"~T.stringof~"() caught error: "~o.toString()).toStringz());
    }
}

/// ゲッターの定義
private extern(C)
SQInteger sqd_classfieldgetter(T)(HSQUIRRELVM sq) {
    try {
        auto self_ = sqd_get!SquirrelObject(sq, 1);
        auto slotname = sqd_get!SQDString(sq, 2);
        auto self = self_.dobject!T;
        
        pragma(msg, "make getter");
        foreach (i, field_; __traits(derivedMembers, T)) {
            enum field = __traits(derivedMembers, T)[i];
            //static if (isMemberVariable!(T, field)) {
            static if (isProperty!(T, field) ||
                       isPropertyGetOnly!(T, field)) {
                pragma(msg, T.stringof~"."~field);
                if (slotname == field) {
                    auto retval =  __traits(getMember, self, field);
                    sqd_push(sq, retval);
                    return 1;
                }
            }
        }
        return sq_throwerror(sq, ("sqd_classfieldgetter!"~T.stringof~"() get undefined slot: "~slotname).toStringz());
    }
    catch (Throwable o) {
        return sq_throwerror(sq, ("sqd_classfieldgetter!"~T.stringof~"() caught error: "~o.toString()).toStringz());
    }
}

/// コンストラクタの定義
private extern(C)
SQInteger sqd_classconstructor(T)(HSQUIRRELVM sq) {
    try {
        auto self_ = sqd_get!SquirrelObject(sq, 1);
        auto obj = sqd_get!SQUserPointer(sq, 2);
        //writeln(self_.toString());
        //writeln(sqd_tostring(sq, 2));
        self_.set(DObjectSlotname, obj);
        return 0;
    }
    catch (Throwable o) {
        return sq_throwerror(sq, ("sqd_classconstructor!"~T.stringof~"() caught error: "~o.toString()).toStringz());
    }
}



/// クラスを定義する。
void sqd_classensuredefine(T)(HSQUIRRELVM sq) {
    mixin(sqd_assert_stack_is_unchanged);
    static bool classdefined=false;
    if (!classdefined) {
        pragma(msg, "make class wraps "~T.stringof);
        sq_newclass(sq, false);
        sqd_newslot(sq, DObjectSlotname, sqd_null);
        // set methods
        foreach (i, field_; __traits(derivedMembers, T)) {
            enum field = __traits(derivedMembers, T)[i];
            static if (field != "__ctor" &&
                       //isMemberFunction!(T, field) &&
                       //isAccessible!(T, field) &&
                       !isProperty!(T, field) &&
                       !isPropertyGetOnly!(T, field) &&
                       __traits(compiles, sqd_pushmethodwrapper!(T, field)(sq))) {
                //mixin printFieldAttribute!(T, field);
                pragma(msg, T.stringof~"."~field~"()");
                sqd_push(sq, field);
                sqd_pushmethodwrapper!(T, field)(sq);
                sqd_checkresult(sq_newslot(sq, -3, SQFalse), sq, "newslot");
            }
        }
        sqd_newslot(sq, "constructor", &sqd_classconstructor!T);
        sqd_newslot(sq, "_get", &sqd_classfieldgetter!T);
        sqd_newslot(sq, "_set", &sqd_classfieldsetter!T);

        sq_pushroottable(sq);
        sqd_push(sq, T.stringof);
        sq_push(sq, -3); // class
        sqd_checkresult(sq_newslot(sq, -3, SQFalse), sq, "newslot");
        sq_pop(sq, 2); // roottable, class

        classdefined = true;
    }
}

// sqd_pushclassで使われる
private extern(C)
SQInteger sqd_classreleasehook(SQUserPointer p, SQInteger size) {
    GC.removeRoot(p);
    return 0;
}


