/**
   粋がって作ってみたものの、だいたい sqstd_seterrorhandlers で代替できます。
 */
module sq4d.debug_;

import sq4d.all;
import std.string;

private {
    TraceLine[] debugtrace_list;
    TraceLine debugtrace_line;
    bool enabled=false;
    bool enabled_line=false;
}

struct TraceLine {
    SQDString source="null";
    int line=-1;
    SQDString funcname="null";
    
    string toString() {
        return format("%s(%s): %s()", source, line, funcname);
    }
}

/**
   蓄積されたデバッグ情報をprintする。
 */
void sqd_debugprint(HSQUIRRELVM sq) {
    if (enabled) {
        auto print = sq_getprintfunc(sq);
        print(sq, "print debugtrace:\n");
        auto ls = debugtrace_line.toString();
        if (ls.length)
            print(sq, "LINE %.*s\n", ls);
        print(sq, "-- CALLS -------------------------------\n");
        foreach_reverse (trace; debugtrace_list) {
            auto ts = trace.toString();
            if (ts.length) {
                print(sq, "  %.*s\n", ts);
            }
        }
        print(sq, "----------------------------------------\n");
    }
}

extern(C) SQInteger myErrorHandler(HSQUIRRELVM sq) {
    auto print = sq_getprintfunc(sq);
    print(sq, "Squirrel error: %.*s\n", sqd_tostring(sq, -1));
    sqd_debugprint(sq);
    throw new Exception("Squirrel error: "~sqd_tostring(sq, -1).idup);
    assert(0);
    //return 0;
}


/**
   登録用関数
 */
extern(C)
SQInteger sqd_debughook(HSQUIRRELVM sq) {
    auto pf = sq_getprintfunc(sq);
    auto event      = sqd_get!int(sq, -4);
    auto sourcefile = sqd_get!SQDString(sq, -3);
    auto line       = sqd_get!int(sq, -2);
    auto funcname   = sqd_get!SQDString(sq, -1);
    //pf(sq, "debughook argc %d\n", sq_gettop(sq));
    if (event == 'l') {
        debugtrace_line = TraceLine(sourcefile, line, funcname);
        //pf(sq, "LINE %.*s@%.*s:%d\n", funcname, sourcefile, line);
    }
    else if (event == 'c') {
        //pf(sq, "CALL %.*s@%.*s:%d\n", funcname, sourcefile, line);
        if (enabled_line)
            debugtrace_list ~= debugtrace_line;
        debugtrace_list ~= TraceLine(sourcefile, line, funcname);
    }
    else if (event == 'r') {
        if (enabled_line)
            debugtrace_list.length = debugtrace_list.length - 2;
        else
            debugtrace_list.length = debugtrace_list.length - 1;
        //pf(sq, "RETURN %.*s@%.*s:%d\n", funcname, sourcefile, line);
    }
    else {
        pf(sq, "DEBUGHOOK error\n");
        //assert(0);
    }
    return 0;
}

/**
   デバッグモードの有効化
 */
void sqd_debugenable(HSQUIRRELVM sq, bool enable_line=true) {
    enabled = true;
    debugtrace_list.length = 1024;
    debugtrace_list.length = 0;
    if (enable_line) {
        enabled_line = true;
        sq_enabledebuginfo(sq, SQTrue);
    }
    sqd_push(sq, &sqd_debughook);
    sq_setdebughook(sq);
}

