
module sq4d.func;

import sq4d.all;
import std.conv;
import std.traits;
import std.string;
import std.exception;
import std.stdio;


/// D関数ラップをしてpushする。
/// この関数はスタックを変動させる。
void sqd_pushfunction(alias F)(HSQUIRRELVM sq) {
    sq_newclosure(sq, &SQFuncWrap!F, 0);
    enum schemer = "." ~ SquirrelSchemer!(ParameterTypeTuple!F) ~ "\0";
    enum schemer_nz = "." ~ SquirrelSchemer!(ParameterTypeTuple!F);
    pragma(msg, ""~(&F).stringof~" schemer: ["~schemer_nz~"]");
    sqd_checkresult(sq_setparamscheck(sq, SQ_MATCHTYPEMASKSTRING, schemer), sq, "sq_setparamscheck("~schemer~")");
}

/// sqd_pushfunction と同じだが、関数が第一引数に環境コンテキスト(this)を受け取ると仮定する。
/// この関数はスタックを変動させる。
void sqd_pushfunction2(alias F)(HSQUIRRELVM sq) {
    sq_newclosure(sq, &SQFuncWrap!F, 0);
    enum schemer = SquirrelSchemer!(ParameterTypeTuple!F) ~ "\0";
    sqd_checkresult(sq_setparamscheck(sq, SQ_MATCHTYPEMASKSTRING, schemer), sq, "sq_setparamscheck("~schemer~")");
}

/// D関数をSquirrelオブジェクトにラップする。
SquirrelObject sqd_wrapfunction(alias F)(HSQUIRRELVM sq) {
    mixin(sqd_assert_stack_is_unchanged);
    sqd_pushfunction!F(sq);
    scope(success) sq_pop(sq, 1);
    return sqd_get!SquirrelObject(sq, -1);
}

/// sqd_wrapfunction と同じだが、 sqd_pushfunction2を使う。
SquirrelObject sqd_wrapfunction2(alias F)(HSQUIRRELVM sq) {
    mixin(sqd_assert_stack_is_unchanged);
    sqd_pushfunction2!F(sq);
    scope(success) sq_pop(sq, 1);
    return sqd_get!SquirrelObject(sq, -1);
}


/**
   Dで使える関数をSquirrel用にラップする。
 */
extern(C)
SQInteger SQFuncWrap(alias F)(HSQUIRRELVM sq) {
    try {
        alias ParameterTypeTuple!F PT;
        alias ReturnType!F RT;
        //argument
        /+
         // replaces to sq_setparamsscheck
         auto argc = sq_gettop(sq);
         if (argc-1 != PT.length) {
             return sq_throwerror(sq, "argc error");
         }
        +/
        /+
         for example
          argc == 3            -> this, a, b
          PT.length == 2       -> a, b
          i = [0, 1]           -> a, b
          stack = [-3, -2, -1] -> this, a, b
          -> index = -PT.length + i;
         +/
        Tuple!(PT) args;
        foreach (i, v; PT) {
            args[i] = sqd_get!(PT[i])(sq, -PT.length + i);
        }
        //call&return
        static if (is(RT == void)) {
            F(args.tupleof);
            return 0;
        }
        else {
            RT r = F(args.tupleof);
            sqd_push(sq, r);
            return 1;
        }
    }
    catch (Throwable o) {
        return sq_throwerror(sq, o.toString().toStringz());
    }
}


/**
   -2の位置の関数を呼び出して返値を返す。-1はself。
   関数とselfはpopされる。
   この関数はスタックを変動させる。
 */
R sqd_call(R=void, T...)(HSQUIRRELVM sq, T p) {
    auto top = sq_gettop(sq);
    sqd_enforce(top >= 2, sq, "sqd_call stack error");
    foreach (i, v; T) {
        sqd_push(sq, p[i]);
    }

    //sqd_print(sq, T.length.stringof~"\n");
    //clsDx();
    //sqd_printstack(sq);
    //WaitKey();
    enum errmsg = "sq_call("~T.stringof~")";
    sqd_checkresult(sq_call(sq, T.length+1, SQTrue, SQTrue), sq, errmsg);
    scope(success) sq_settop(sq, top-2); // self, closure
    static if (is(R == void))
        return;
    else
        return sqd_get!R(sq, -1);
}

/// 引数チェック用識別子を作成する。
template SquirrelSchemer(T) {
    static if (is(T == int) || is(T == uint)) {
        enum SquirrelSchemer = "i";
    }
    else static if (is(T == float)) {
        enum SquirrelSchemer = "f";
    }
    else static if (is(T == double) || is(T == real)) {
        enum SquirrelSchemer = "n";
    }
    else static if (is(T == string) || is(T == SQDString)) {
        enum SquirrelSchemer = "s";
    }
    else static if (is(T == bool)) {
        enum SquirrelSchemer = "b";
    }
    else static if (is(T == SQUserPointer)) {
        enum SquirrelSchemer = "p";
    }
    else static if (is(T == SquirrelObject)) {
        enum SquirrelSchemer = ".";
    }
    else {
        static assert(0, "unimplemented type: SquirrelSchemer!"~T.stringof);
    }
}
/// ditto
template SquirrelSchemer(T, R...) {
    enum SquirrelSchemer = SquirrelSchemer!T ~ SquirrelSchemer!R;
}
/// ditto
template SquirrelSchemer() {
    enum SquirrelSchemer = "";
}





