
module sq4d.template_;

/// fieldメンバがTのメンバ関数ならtrue
template isMemberFunction(T, string field) {
    static if (isMember!(T, field)) {
        static if (!mixin("__traits(isStaticFunction, T." ~ field ~ ")")) {
            enum isMemberFunction =
                mixin("is(typeof(&T.init." ~ field ~ ") == delegate)");
        }
        else {
            enum isMemberFunction = false;
        }
    }
    else {
        enum isMemberFunction = false;
    }
}

/// fieldメンバがTの静的メンバ関数ならtrue
template isStaticFunction(T, string field) {
    static if (mixin("is(typeof(&T.init." ~ field ~ ") == delegate)")) {
        enum isStaticFunction =
            mixin("__traits(isStaticFunction, T." ~ field ~ ")");
    }
    else {
        enum isStaticFunction = false;
    }
}

/// fieldメンバがTの静的メンバならtrue。・・・多分違う
template isStatic(T, string field) {
    enum isStatic = mixin("__traits(compiles, &T." ~ field ~ ")");
}

/// fieldメンバはTのアクセス可能メンバならtrue。
/// 型やprivateはアクセス不可能なのでfalse。
template isAccessible(T, string field) {
    enum isAccessible =
        mixin("__traits(compiles, T.init." ~ field ~ ")") &&
        mixin("__traits(compiles, &T.init." ~ field ~ ")") &&
        mixin("__traits(compiles, T." ~ field ~ ")");
}

/// fieldメンバがTの動的メンバならtrue
template isMember(T, string field) {
    enum isMember = mixin("__traits(compiles, &T.init." ~ field ~ ")");
}

/// fieldメンバがTの動的メンバ変数ならtrue
template isMemberVariable(T, string field) {
    enum isMemberVariable = isMember!(T, field) && !isMemberFunction!(T, field) && !isStaticFunction!(T, field);
}

/// 読み書き可能なプロパティや、publicメンバ変数ならtrue。
template isProperty(T, string field) {
    enum isProperty = __traits(compiles, mixin("T.init."~field~" = T.init."~field));
}

/// 読み取りのみのプロパティならtrue。
template isPropertyGetOnly(T, string field) {
    static if (__traits(compiles, mixin("is(typeof(T.init."~field~") == typeof(T.init."~field~"()))")))
        enum isPropertyGetOnly = mixin("is(typeof(T.init."~field~") == typeof(T.init."~field~"()))");
    else
        enum isPropertyGetOnly = false;
}

/// fieldメンバの属性をコンパイル時表示。mixin用テンプレート
template PrintFieldAttribute(T, string field) {
    pragma(msg, T.stringof~"."~field);
    pragma(msg, "isAccessible: "~isAccessible!(T, field).stringof);
    pragma(msg, "isMember: "~isMember!(T, field).stringof);
    pragma(msg, "isMemberFunction: "~isMemberFunction!(T, field).stringof);
    pragma(msg, "isStaticFunction: "~isStaticFunction!(T, field).stringof);
}

