/**
   std.typecons.tuple をSquirrelで使うための操作
 */
module sq4d.tuple;

import sq4d.all;
public import std.typecons;
import std.exception;

/// Dのタプル構造体をSquirrel配列に変換してpush
/// この関数をスタックを変動させる。
void sqd_pushtuple(T)(HSQUIRRELVM sq, auto ref T t)
    if (isTuple!T)
{
    enum len = t.tupleof.length;
    sq_newarray(sq, len);
    foreach (i, v; t.tupleof) {
        sqd_push(sq, i);
        sqd_push(sq, t[i]);
        sqd_checkresult(sq_set(sq, -3), sq, "sqd_pushtuple");
    }
}

/// Squirrel配列をDのタプル構造体に変換
T sqd_gettuple(T)(HSQUIRRELVM sq, int index)
    if (isTuple!T)
{
    mixin(sqd_assert_stack_is_unchanged);
    sqd_enforce(sq_gettype(sq, index) == OT_ARRAY, sq, "sqd_gettuple() wrong type");
    sqd_enforce(sq_getsize(sq, index) == T.tupleof.length, sq, "sqd_gettuple() wrong array size");
    T ret;
    foreach (i, v; ret.tupleof) {
        sqd_push(sq, i);
        sqd_checkresult(sq_get(sq, -2), sq, "sqd_pushtuple");
        ret[i] = sqd_get!(typeof(ret[i]))(sq, -1);
        sq_pop(sq, 1);
    }
    return ret;
}





