/* sq.c for D */
/*	see copyright notice in squirrel.h */

import std.c.stdio;
import std.c.stdlib;
import std.c.string;
import std.c.stdarg;
import squirrel;
import sqstdblob;
import sqstdsystem;
import sqstdio;
import sqstdmath;
import sqstdstring;
import sqstdaux;
import core.stdc.wchar_;
import std.string;


version (SQUNICODE) {
    alias fwprintf scfprintf;
    //alias _wfopen scfopen;
    alias vfwprintf scvprintf;
}
else {
    alias fprintf scfprintf;
    //alias fopen scfopen;
    alias vfprintf scvprintf;
}

/+
#if defined(_MSC_VER) && defined(_DEBUG)
int MemAllocHook( int allocType, void *userData, size_t size, int blockType, 
   long requestNumber, const unsigned char *filename, int lineNumber)
{
	//if(requestNumber==769)_asm int 3;
	return 1;
}
#endif
+/

extern(C) SQInteger quit(HSQUIRRELVM v) {
    int* done;
    sq_getuserpointer(v, -1, cast(SQUserPointer*)&done);
    *done=1;
    return 0;
}

extern(C) void printfunc(HSQUIRRELVM v, const(SQChar)* s, ...)
{
    va_list vl;
    va_start(vl, s);
    scvprintf(stdout, s, vl);
    va_end(vl);
}

extern(C) void errorfunc(HSQUIRRELVM v,const(SQChar)*s, ...)
{
    va_list vl;
    va_start(vl, s);
    scvprintf(stderr, s, vl);
    va_end(vl);
}

void PrintVersionInfos()
{
    scfprintf(stdout, "%.*s %.*s (%d bits)\n", SQUIRREL_VERSION, SQUIRREL_COPYRIGHT, SQInteger.sizeof*8);
}

void PrintUsage()
{
    scfprintf(stderr,
              "usage: sq <options> <scriptpath [args]>.\n"
              "Available options are:\n"
              "   -c              compiles the file to bytecode(default output 'out.cnut')\n"
              "   -o              specifies output file for the -c option\n"
              "   -c              compiles only\n"
              "   -d              generates debug infos\n"
              "   -v              displays version infos\n"
              "   -h              prints help\n");
}

enum _INTERACTIVE = 0;
enum _DONE        = 2;
enum _ERROR       = 3;
//<<FIXME>> this func is a mess
int getargs(HSQUIRRELVM v, string[] args, SQInteger* retval)
{
	int i;
	int compiles_only;
	static SQChar[500] temp;
	const(SQChar)* ret;
	string output;
	int lineinfo;
	*retval = 0;
	if(args.length > 1) {
		int arg=1, exitloop=0;
		
		while(arg < args.length && !exitloop) {
            if(args[arg][0]=='-') {
				switch(args[arg][1]) {
				case 'd': //DEBUG(debug infos)
					sq_enabledebuginfo(v,1);
					break;
				case 'c':
					compiles_only = 1;
					break;
				case 'o':
					if(arg < args.length) {
						arg++;
						output = args[arg];
					}
					break;
				case 'v':
					PrintVersionInfos();
					return _DONE;
				
				case 'h':
					PrintVersionInfos();
					PrintUsage();
					return _DONE;
				default:
					PrintVersionInfos();
					scprintf("unknown prameter '-%c'\n", args[arg][1]);
					PrintUsage();
					*retval = -1;
					return _ERROR;
				}
			}else break;
			arg++;
		}

		// src file
		
		if(arg<args.length) {
			const(SQChar)* filename;
            version (SQUNICODE) {
                mbstowcs(temp.ptr, args[arg].ptr, args[arg].length);
                filename=temp.ptr;
            }
            else {
                filename=args[arg].toStringz();
            }

			arg++;
			
			//sq_pushstring(v,_SC("ARGS"),-1);
			//sq_newarray(v,0);
			
			//sq_createslot(v,-3);
			//sq_pop(v,1);
			if(compiles_only) {
				if(SQ_SUCCEEDED(sqstd_loadfile(v,filename,SQTrue))){
					const(SQChar)* outfile = "out.cnut";
					if(output) {
                        version (SQUNICODE) {
                            int len = cast(int)(output.length+1);
                            mbstowcs(sq_getscratchpad(v,len*SQChar.sizeof), output.ptr, len);
                            outfile = sq_getscratchpad(v,-1);
                        }
                        else {
                            outfile = output.toStringz();
                        }
					}
					if(SQ_SUCCEEDED(sqstd_writeclosuretofile(v,outfile)))
						return _DONE;
				}
			}
			else {
				//if(SQ_SUCCEEDED(sqstd_dofile(v,filename,SQFalse,SQTrue))) {
					//return _DONE;
				//}
				if(SQ_SUCCEEDED(sqstd_loadfile(v,filename,SQTrue))) {
					int callargs = 1;
					sq_pushroottable(v);
					for(i=arg;i<args.length;i++)
					{
						const(SQChar)* a;
                        version (SQUNICODE) {
                            int alen=cast(int)args[i].length;
                            a=sq_getscratchpad(v,cast(int)(alen*SQChar.sizeof));
                            mbstowcs(sq_getscratchpad(v,-1), args[i].ptr, alen);
                            sq_getscratchpad(v,-1)[alen] = '\0';
                        }
                        else {
                            a=args[i].toStringz();
                        }
						sq_pushstring(v,a,-1);
						callargs++;
						//sq_arrayappend(v,-2);
					}
					if(SQ_SUCCEEDED(sq_call(v,callargs,SQTrue,SQTrue))) {
						SQObjectType type = sq_gettype(v,-1);
						if (type == OT_INTEGER) {
							*retval = type;
							sq_getinteger(v,-1,retval);
						}
						return _DONE;
					}
					else{
						return _ERROR;
					}
					
				}
			}
			//if this point is reached an error occured
			{
				const(SQChar)* err;
				sq_getlasterror(v);
				if(SQ_SUCCEEDED(sq_getstring(v,-1,&err))) {
					scprintf("Error [%s]\n", err);
					*retval = -2;
					return _ERROR;
				}
			}
			
		}
	}

	return _INTERACTIVE;
}

void Interactive(HSQUIRRELVM v)
{
	
    enum MAXINPUT = 1024;
	SQChar buffer[MAXINPUT];
	SQInteger blocks =0;
	SQInteger string=0;
	SQInteger retval=0;
	SQInteger done=0;
	PrintVersionInfos();
		
	sq_pushroottable(v);
	sq_pushstring(v,"quit",-1);
	sq_pushuserpointer(v,&done);
	sq_newclosure(v,&quit,1);
	sq_setparamscheck(v,1,null);
	sq_createslot(v,-3);
	sq_pop(v,1);

    while (!done) 
	{
		SQInteger i = 0;
		scprintf("\nsq>");
		for(;;) {
			int c;
			if(done)return;
			c = getchar();
			if (c == '\n') {
				if (i>0 && buffer[i-1] == '\\')
				{
					buffer[i-1] = '\n';
				}
				else if(blocks==0)break;
				buffer[i++] = '\n';
			}
			else if (c=='}') {blocks--; buffer[i++] = cast(SQChar)c;}
			else if(c=='{' && !string){
					blocks++;
					buffer[i++] = cast(SQChar)c;
			}
			else if(c=='"' || c=='\''){
					string=!string;
					buffer[i++] = cast(SQChar)c;
			}
			else if (i >= MAXINPUT-1) {
				scfprintf(stderr, "sq : input line too long\n");
				break;
			}
			else{
				buffer[i++] = cast(SQChar)c;
			}
		}
		buffer[i] = '\0';
		
		if(buffer[0]=='='){
			scsprintf(sq_getscratchpad(v,MAXINPUT), "return (%s)", &buffer[1]);
			memcpy(buffer.ptr, sq_getscratchpad(v,-1), (scstrlen(sq_getscratchpad(v,-1))+1)*SQChar.sizeof);
			retval=1;
		}
		i=scstrlen(buffer.ptr);
		if(i>0){
			SQInteger oldtop=sq_gettop(v);
			if(SQ_SUCCEEDED(sq_compilebuffer(v,buffer.ptr,i,"interactive console",SQTrue))){
				sq_pushroottable(v);
				if(SQ_SUCCEEDED(sq_call(v,1,retval,SQTrue)) &&	retval){
					scprintf("\n");
					sq_pushroottable(v);
					sq_pushstring(v,"print",-1);
					sq_get(v,-2);
					sq_pushroottable(v);
					sq_push(v,-4);
					sq_call(v,2,SQFalse,SQTrue);
					retval=0;
					scprintf("\n");
				}
			}
			
			sq_settop(v,oldtop);
		}
	}
}

int main(string[] args)
{
	HSQUIRRELVM v;
	SQInteger retval = 0;
	const(SQChar)* filename;
/+#if defined(_MSC_VER) && defined(_DEBUG)
	_CrtSetAllocHook(MemAllocHook);
#endif
+/
	
	v=sq_open(1024);
	sq_setprintfunc(v,&printfunc,&errorfunc);

	sq_pushroottable(v);

	sqstd_register_bloblib(v);
	sqstd_register_iolib(v);
	sqstd_register_systemlib(v);
	sqstd_register_mathlib(v);
	sqstd_register_stringlib(v);

	//aux library
	//sets error handlers
	sqstd_seterrorhandlers(v);

	//gets arguments
	switch(getargs(v,args,&retval))
	{
	case _INTERACTIVE:
		Interactive(v);
		break;
	case _DONE:
	case _ERROR:
	default: 
		break;
	}

	sq_close(v);
	
/+#if defined(_MSC_VER) && defined(_DEBUG)
	_getch();
	_CrtMemDumpAllObjectsSince( NULL );
#endif
+/
	return retval;
}

